
#ifndef STDIO_INCLUDED
#include <stdio.h>
#define STDIO_INCLUDED
#endif

#ifndef PORT_INCLUDED
#include "port.h"
#define PORT_INCLUDED
#endif

#ifndef LPTEK_INCLUDED
#include "lptek.h"
#define LPTEK_INCLUDED
#endif



inline unsigned char reconstruct(unsigned char weirdHigh, unsigned char weirdLow)

// This routine is necessary to reconstruct a full byte from two 4-bit
// nibbles sent sequentially into the computer's parallel port.  The format
// is very strange; see LPTek documentation for details.  Each nibble is
     // isolated, split, and reassembled, before being OR'd with the other
     // nibble to produce the result.

{
  // 112 = 0111 0000
  // 7 = 0000 0111
  // 240 = 1111 0000
  register unsigned char lowChar = ( (7 & (weirdLow>>3)) | (16 & (weirdLow>>3)^16)>>1 );
  register unsigned char highChar = ( (112 & (weirdHigh<<1)) | 
				      (128 & (weirdHigh ^128) ) );
  return (lowChar | highChar);
}


inline void sendBits(const char& chip, const char& buffer, 
		     const unsigned char& bitChar)

     // send binary output, one byte, to the LP48IO board, to a chip and 
     //buffer which  has already been configured for output

{
  // board 1 = address 16 (upper nibble)
  // chip 1 = 000-010; chip 1 = 100-110 (011, 111 = control registers)
  port_out(0x0378, BITSBOARD + 4*chip + buffer  )  ; // address of buffer 
  port_out(0x037A, 2 )  ;  // addr latch
  port_out(0x037A, 7 )  ;  // confirm/preset
  
  port_out(0x0378, bitChar )  ; // present data 
  port_out(0x037A, 0 )  ;  // write command
  port_out(0x037A, 7 )  ;  // confirm/preset
}


inline unsigned char getBits(const char& chip, const char& buffer)

     //retrieve a byte of binary data from a particular chip and buffer
     // on the LP48IO board, assuming that chip/buffer has already 
     // been configured for input.

{
  unsigned char lowChar;
  // board 1 = address 16 (upper nibble)
  // chip 1 = 0000-0010; chip 1 = 0100-0110 (0011, 0111 = control registers)
  port_out(0x0378, BITSBOARD + 4*chip + buffer  )  ; // address
  port_out(0x037A, 2 )  ;  // addr latch
  port_out(0x037A, 7 )  ;  // confirm/preset
  
  
  port_out(0x037A, 1 )  ;  // read instruction
  port_out(0x037A, 7 )  ;  // confirm/preset  
  lowChar = port_in(0x0379); //read in low nibble
  port_out(0x037A, 0xF )  ;  // command read high
  return reconstruct( port_in(0x0379), lowChar);  // assemble full byte
}


inline void setBitsControl(const char& chip, char type)

     // send control bits to a chip on LP48IO board, telling it 
     // which of three buffers to use as input and/or output
     // This routine is presently configured only for all three
     // buffers as either input or output; all other permutations
     // are available in the chip itself (by sending control instructions
     // other than 128 or 155); look for those instructions on page 4-8 of
     // the 82C55A documenation (available online on the Harris Semiconductor
     // chip-data search page, or elsewhere).
{

  // sending control word to fourth buffer of chip (= control register;
  // the first three registers are i/o registers A,B,C)
  
  //output
  if (type == 'o')
    sendBits(chip, 3,128);

  //input
  if (type == 'i')
    sendBits(chip, 3, 155);
}


showBits (const unsigned char inChar)  
     // this function prints out the bit-field, 1's and 0's, of its argument
     // as a diagnostic
{
  unsigned char uch;
  int power;
  for (int ii = 7; ii >=0; ii--){
    uch = 1;
    for (int power = 0; power < ii; power++)
      uch *= 2;
    if (inChar & uch)
      printf("%d",1);
    else
      printf("%d",0);

  }
  printf ("\n");
}

inline unsigned char getADC(const int& channel)

     //retrieves the a byte of (formerly) analog voltage from the LP8AD8 board.
     //This form is unusually long and complex, in order to retrieve data
     // instantly from an arbitrary channel. It can be sped up if you will 
     // settle for retrieving data from the same address time after time,
     // and/or if you will settle for retrieving the data stored at the
     // previous read instruction rather than the current one.
{

  unsigned char lowChar;

  // send and latch the board's address
  port_out(0x037A, 7 )  ;  // confirm/preset
  port_out(0x0378, ADCBOARD )  ; // address of board 2
  port_out(0x037A, 2 )  ;  // addr latch
  port_out(0x037A, 7 )  ;  // confirm/preset

  // send and "write" the channel's location on that board
  // (for some odd reason, the board and channel information have
  // to be send separately, with distinct commands...not the usual LPTek
  // format).
  port_out(0x0378, channel )  ; // address of channel
  port_out(0x037A, 0 )  ;  //  write
  port_out(0x037A, 7 )  ;  // confirm/preset
  
  // send and "read" the  two data nibbles  NOTE: an undocumented 
  // "feature" of this board seems to be that you normally read out 
  // the *previously* digitized byte from a FIFO, which means you need 
  // to actually make two read cycles to get the current one.  

  // First read: don't keep input data
  port_out(0x037A, 1 )  ;  // read instruction
  port_out(0x037A, 7 )  ;  // confirm/preset  
  port_in(0x0379);          // retrieve low nibble through port
  port_out(0x037A, 0xF )  ;  // read high
  port_in(0x0379);            // retrieve high nibble through port
 
  // second read: this one for keeps

  port_out(0x0378, channel )  ; // address of channel
  port_out(0x037A, 0 )  ;  //  write
  port_out(0x037A, 7 )  ;  // confirm/preset

  port_out(0x037A, 1 )  ;  // read (low) instruction
  port_out(0x037A, 7 )  ;  // confirm/preset  
  lowChar = port_in(0x0379);  // FINALLY get the lower nibble
  port_out(0x037A, 0xF )  ;  // read high instruction

  return reconstruct( port_in(0x0379), lowChar);
 }




inline unsigned char* getAllADC(unsigned char* outArray)

     //retrieves all eight byted of (formerly) analog voltage from the LP8AD8
     // board.
     // Assumes outArray is dimensioned to [8]
{

  unsigned char lowChar;

  // send and latch the board's address just once
  port_out(0x037A, 7 )  ;  // confirm/preset
  port_out(0x0378, ADCBOARD )  ; // address of board 2
  port_out(0x037A, 2 )  ;  // addr latch
  port_out(0x037A, 7 )  ;  // confirm/preset

  for (int channel = 0; channel < 9; channel ++){
   
    // send and "write" the channel's location on that board
    // (for some odd reason, the board and channel information have
    // to be send separately, with distinct commands...not the usual LPTek
    // format).
    port_out(0x0378, channel )  ; // address of channel
    port_out(0x037A, 0 )  ;  //  write
    port_out(0x037A, 7 )  ;  // confirm/preset
  
    // send and "read" the  two data nibbles  NOTE: an undocumented 
    // "feature" of this board seems to be that you normally read out 
    // the *previously* digitized byte from a FIFO, which means you are
    // actually reading in the previosly digitized channel.

    port_out(0x0378, channel )  ; // address of channel
    port_out(0x037A, 0 )  ;  //  write
    port_out(0x037A, 7 )  ;  // confirm/preset

    port_out(0x037A, 1 )  ;  // read (low) instruction
    port_out(0x037A, 7 )  ;  // confirm/preset  
    lowChar = port_in(0x0379);  // FINALLY get the lower nibble
    port_out(0x037A, 0xF )  ;  // read high instruction

    // store the presently read channel into the previos buffer location,
    // to compensate for the buffer delay on the board.
    if (channel > 0)
      outArray[channel-1] = reconstruct( port_in(0x0379), lowChar);
  }
  return outArray;
 }


inline void sendDAC(const unsigned char& channel, const unsigned char& data)

     // send a byte of (soon-to-be) analog data to the LP8DA8 board
{

  // An undocumented "feature" seems to be that this board is dyslexic 
  // (lysdexic?), using left-right reversed bytes.  So you need to
  //disentangle data (reverse bit order) before sending.

  register unsigned char reversedData;
  reversedData =( (data&128)>>7 )|( (data&64)>>5) | ((data&32)>>3) | 
    ((data&16)>>1) | ((data&8)<<1) | ((data&4)<<3) | ((data&2)<<5) |
    ((data&1)<<7) ;

   port_out(0x0378, DACBOARD + channel )  ; // address of board/channel
   port_out(0x037A, 7 )  ;  //  preset
   port_out(0x037A, 2 )  ;  // addr latch
   port_out(0x037A, 7 )  ;  // confirm/preset

   port_out(0x0378, reversedData )  ; // data 
   port_out(0x037A, 0 )  ;  // write command
   port_out(0x037A, 7 )  ;  // confirm/preset

   // another oddity: even after sending the data, you need to send
   // a "read" command to tell the board to use it.
   port_out(0x037A, 1 )  ;  // 'read' necessary to latch output
   port_out(0x037A, 7 )  ;  // confirm/preset

}



inline void sendAllDAC( unsigned char* dataArray)

     // send 8 bytes of (soon-to-be) analog data to the LP8DA8 board
{

  // An undocumented "feature" seems to be that this board is dyslexic 
  // (lysdexic?), using left-right reversed bytes.  So you need to
  //disentangle data (reverse bit order) before sending.

  register unsigned char reversedData, data;
  for (int channel=0; channel < 8; channel++){
    data = dataArray[channel];
    reversedData =( (data&128)>>7 )|( (data&64)>>5) | ((data&32)>>3) | 
      ((data&16)>>1) | ((data&8)<<1) | ((data&4)<<3) | ((data&2)<<5) |
      ((data&1)<<7) ;

    port_out(0x0378, DACBOARD + channel )  ; // address of board/channel
    port_out(0x037A, 7 )  ;  //  preset
    port_out(0x037A, 2 )  ;  // addr latch
    port_out(0x037A, 7 )  ;  // confirm/preset

    port_out(0x0378, reversedData )  ; // data 
    port_out(0x037A, 0 )  ;  // write command
    port_out(0x037A, 7 )  ;  // confirm/preset

    // another oddity: even after sending the data, you need to send
    // a "read" command to tell the board to use it.
    port_out(0x037A, 1 )  ;  // 'read' necessary to latch output
    port_out(0x037A, 7 )  ;  // confirm/preset
  }
}






